# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import asyncio
import re
import traceback

# for concurrent processing
from contextlib import suppress
from typing import Dict, List, Optional

import aiohttp
from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.facet import ErrorMessageRunFacet
from openlineage.common.provider.dbt import DbtCloudArtifactProcessor, ParentRunMetadata

from airflow.hooks.base import BaseHook
from airflow.providers.dbt.cloud.hooks.dbt import (
    DbtCloudHook,
    fallback_to_default_account,
)


class DbtCloudExtractor(BaseExtractor):
    default_schema = "public"

    def __init__(self, operator):
        super().__init__(operator)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return [
            "DbtCloudRunJobOperator",
            "DbtCloudJobRunSensor",
            "DbtCloudRunJobOperatorAsync",
            "DbtCloudJobRunSensorAsync",
        ]

    def get_task_metadata(self):
        try:
            operator = self.operator
            task_name = f"{operator.dag_id}.{operator.task_id}"
            run_facets: Dict = {}
            job_facets: Dict = {}

        except Exception as e:
            self.log.exception("Exception has occurred in extract()")
            error = ErrorMessageRunFacet(str(e), "python")
            error.stackTrace = traceback.format_exc()
            run_facets["errorMessage"] = error
            raise Exception(e)
        finally:
            # finally return the task metadata
            return TaskMetadata(
                name=task_name,
                inputs=[],
                outputs=[],
                run_facets=run_facets,
                job_facets=job_facets,
            )

    def extract(self) -> TaskMetadata:
        operator = self.operator
        # according to the job type, pre-perform the necessary
        # fetch (job and project) of data so that it will save
        # time when creating lineage data at the completion
        if operator.task_type in [
            "DbtCloudRunJobOperator",
            "DbtCloudRunJobOperatorAsync",
        ]:
            job_id = operator.job_id
            hook = DbtCloudHook(operator.dbt_cloud_conn_id)
            # use job_id to get connection
            job = hook.get_job(job_id=job_id).json()["data"]
            project_id = job["project_id"]
            project = hook.get_project(project_id=project_id).json()["data"]
            connection = project["connection"]
            self.context["connection"] = connection
            self.context["job"] = job

        elif operator.task_type in [
            "DbtCloudJobRunSensor",
            "DbtCloudJobRunSensorAsync",
        ]:
            run_id = operator.run_id
            connection = BaseHook.get_connection(operator.dbt_cloud_conn_id)
            account_id = connection.login
            hook = DbtCloudHook(operator.dbt_cloud_conn_id)
            job_run = hook.get_job_run(account_id=account_id, run_id=run_id, include_related=["job"]).json()[
                "data"
            ]
            project_id = job_run["project_id"]
            job = job_run["job"]
            project = hook.get_project(project_id=project_id, account_id=account_id).json()["data"]
            connection = project["connection"]
            self.context["connection"] = connection
            self.context["job"] = job

        return self.get_task_metadata()

    # internal method to extract dbt lineage and send it to
    # OL backend
    def extract_dbt_lineage(self, operator, run_id, job_name=None):
        from openlineage.airflow.adapter import (
            _DAG_NAMESPACE,
            _PRODUCER,
            OpenLineageAdapter,
        )

        job = self.context["job"]
        account_id = job["account_id"]
        hook = DbtCloudHook(operator.dbt_cloud_conn_id)
        execute_steps = job["execute_steps"]
        job_run = hook.get_job_run(
            run_id=run_id, account_id=account_id, include_related=["run_steps"]
        ).json()["data"]
        run_steps = job_run["run_steps"]
        connection = self.context["connection"]
        steps = []

        for run_step in run_steps:
            name = run_step["name"]
            if name.startswith("Invoke dbt with `"):
                regex_pattern = "Invoke dbt with `([^`.]*)`"
                m = re.search(regex_pattern, name)
                command = m.group(1)
                if command in execute_steps:
                    steps.append(run_step["index"])

        headers, tenant = DbtCloudExtractor.get_headers_tenants_from_connection(hook)

        catalog = None
        with suppress(Exception):
            catalog = hook.get_job_run_artifact(run_id, path="catalog.json").json()["data"]

        async def get_steps(steps):
            tasks = [
                self.get_dbt_artifacts(
                    hook,
                    run_id,
                    account_id,
                    step,
                    ["manifest", "run_results"],
                    headers,
                    tenant,
                )
                for step in steps
            ]
            return await asyncio.gather(*tasks)

        loop = asyncio.new_event_loop()
        step_artifacts = loop.run_until_complete(get_steps(steps))
        for artifacts in step_artifacts:
            # process manifest
            manifest = artifacts["manifest"]
            run_reason = manifest["metadata"]["env"]["DBT_CLOUD_RUN_REASON"]
            # ex: Triggered via Apache Airflow by task 'trigger_job_run1' in the astronomy DAG.
            # regex pattern:
            #   Triggered via Apache Airflow by task '[a-zA-Z0-9_]+' in the [a-zA-Z-0-9_]+ DAG.
            regex_pattern = (
                "Triggered via Apache Airflow by task '([a-zA-Z0-9_]+)' " "in the ([a-zA-Z0-9_]+) DAG\\."
            )
            m = re.search(regex_pattern, run_reason)
            task_id = m.group(1)
            dag_id = m.group(2)

            if not artifacts.get("run_results", None):
                continue

            processor = DbtCloudArtifactProcessor(
                producer=_PRODUCER,
                job_namespace=_DAG_NAMESPACE,
                skip_errors=False,
                logger=self.log,
                manifest=manifest,
                run_result=artifacts["run_results"],
                profile=connection,
                catalog=catalog,
            )

            # parent run is an Airflow task triggering DBT Cloud run
            parent_run_id = self.context["task_uuid"]

            # if parent run id exists, set it as parent run metadata - so that
            # the DBT job can contain parent information (which is the current task)
            if parent_run_id is not None:
                parent_job = ParentRunMetadata(
                    run_id=parent_run_id,
                    job_name=job_name if job_name is not None else f"{dag_id}.{task_id}",
                    job_namespace=_DAG_NAMESPACE,
                )
                processor.dbt_run_metadata = parent_job

            events = processor.parse().events()
            client = OpenLineageAdapter.get_or_create_openlineage_client()
            for event in events:
                client.emit(event)

    async def get_dbt_artifacts(self, hook, run_id, account_id, step, artifacts, headers, tenant):
        async with aiohttp.ClientSession(headers=headers) as session:
            tasks = {
                artifact: DbtCloudExtractor.get_job_run_artifact(
                    hook=hook,
                    session=session,
                    tenant=tenant,
                    run_id=run_id,
                    path=artifact,
                    account_id=account_id,
                    step=step,
                )
                for artifact in artifacts
            }
            results = await asyncio.gather(*tasks.values())
        return {filename: result for filename, result in zip(tasks.keys(), results)}

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        task_meta_data = self.get_task_metadata()
        operator = self.operator
        if (
            operator.task_type in ["DbtCloudRunJobOperator", "DbtCloudRunJobOperatorAsync"]
            and operator.wait_for_termination is True
        ):
            run_id = task_instance.xcom_pull(task_ids=task_instance.task_id, key="return_value")
            if not run_id:
                run_url = task_instance.xcom_pull(task_ids=task_instance.task_id, key="job_run_url")
                run_id = run_url.split("/")[-2]
        elif operator.task_type in [
            "DbtCloudJobRunSensor",
            "DbtCloudJobRunSensorAsync",
        ]:
            run_id = operator.run_id

        self.extract_dbt_lineage(
            operator=operator,
            run_id=run_id,
            job_name=f"{task_instance.dag_id}.{task_instance.task_id}",
        )

        return task_meta_data

    @staticmethod
    def get_headers_tenants_from_connection(hook):
        """Get Headers, tenants from the connection details"""
        headers = {}
        connection = hook.get_connection(hook.dbt_cloud_conn_id)
        tenant = connection.host or "cloud.getdbt.com"
        headers["Content-Type"] = "application/json"
        headers["Authorization"] = f"Token {connection.password}"
        return headers, tenant

    @staticmethod
    @fallback_to_default_account
    async def get_job_run_artifact(hook, session, tenant, run_id: int, path: str, account_id=None, step=None):
        endpoint = f"{account_id}/runs/{run_id}/artifacts/{path}.json"
        url = f"https://{tenant}/api/v2/accounts/{endpoint or ''}"
        async with session.get(url, params={"step": step}) as response:
            try:
                response.raise_for_status()
                return await response.json(content_type=None)
            except aiohttp.ClientResponseError as e:
                raise Exception(str(e.status) + ":" + e.message)



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.run import Dataset


def generate_s3_dataset(path) -> Dataset:
    return Dataset(
        namespace="s3://{}".format(path.replace("s3://", "").split("/")[0]),
        name=path,
        facets={},
    )


class SageMakerProcessingExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["SageMakerProcessingOperator", "SageMakerProcessingOperatorAsync"]

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        inputs = []
        outputs = []

        try:
            inputs, outputs = self._get_s3_datasets(
                processing_inputs=xcom_values["Processing"]["ProcessingInputs"],
                processing_outputs=xcom_values["Processing"]["ProcessingOutputConfig"]["Outputs"],
            )
        except KeyError as e:
            self.log.error(f"Could not find input/output information in Xcom. {e}")

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=outputs,
        )

    def _get_s3_datasets(self, processing_inputs, processing_outputs):
        inputs = []
        outputs = []

        try:
            for processing_input in processing_inputs:
                inputs.append(generate_s3_dataset(processing_input["S3Input"]["S3Uri"]))
        except Exception as e:
            self.log.error(f"Cannot find S3 input details. {e}", exc_info=True)

        try:
            for processing_output in processing_outputs:
                outputs.append(generate_s3_dataset(processing_output["S3Output"]["S3Uri"]))
        except Exception as e:
            self.log.error(f"Cannot find S3 output details. {e}", exc_info=True)

        return inputs, outputs


class SageMakerTransformExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["SageMakerTransformOperator", "SageMakerTransformOperatorAsync"]

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        self.log.debug(f"extract_on_complete({task_instance})")

        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        model_package_arn = None
        transform_input = None
        transform_output = None

        try:
            model_package_arn = xcom_values["Model"]["PrimaryContainer"]["ModelPackageName"]
        except KeyError as e:
            self.log.error(f"Cannot find Model Package Name in Xcom values. {e}", exc_info=True)

        try:
            transform = xcom_values["Transform"]
            transform_input = transform["TransformInput"]["DataSource"]["S3DataSource"]["S3Uri"]
            transform_output = transform["TransformOutput"]["S3OutputPath"]
        except KeyError as e:
            self.log.error(
                f"Cannot find some required input/output details in XCom. {e}",
                exc_info=True,
            )

        inputs = []

        if transform_input is not None:
            inputs.append(generate_s3_dataset(transform_input))

        if model_package_arn is not None:
            model_data_urls = self._get_model_data_urls(model_package_arn)
            for model_data_url in model_data_urls:
                inputs.append(generate_s3_dataset(model_data_url))

        output = []
        if transform_output is not None:
            output.append(generate_s3_dataset(transform_output))

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=output,
        )

    def _get_model_data_urls(self, model_package_arn):
        model_data_urls = []
        try:
            model_containers = self.operator.hook.get_conn().describe_model_package(
                ModelPackageName=model_package_arn
            )["InferenceSpecification"]["Containers"]

            for container in model_containers:
                model_data_urls.append(container["ModelDataUrl"])
        except Exception as e:
            self.log.error(f"Cannot retrieve model details. {e}", exc_info=True)

        return model_data_urls


class SageMakerTrainingExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["SageMakerTrainingOperator", "SageMakerTrainingOperatorAsync"]

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        self.log.debug(f"extract_on_complete({task_instance})")

        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        inputs = []
        output = []

        try:
            for input_data in xcom_values["Training"]["InputDataConfig"]:
                inputs.append(generate_s3_dataset(input_data["DataSource"]["S3DataSource"]["S3Uri"]))
        except KeyError as e:
            self.log.error(f"Issues extracting inputs. {e}")

        try:
            output.append(generate_s3_dataset(xcom_values["Training"]["ModelArtifacts"]["S3ModelArtifacts"]))
        except KeyError as e:
            self.log.error(f"Issues extracting inputs. {e}")

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=output,
        )



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional
from urllib.parse import urlparse

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.extractors.sql_extractor import SqlExtractor
from openlineage.client.facet import SqlJobFacet
from openlineage.common.dataset import Dataset, Source
from openlineage.common.models import DbColumn, DbTableSchema
from openlineage.common.sql import DbTableMeta, SqlMeta, parse


class AthenaExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["AthenaOperator", "AWSAthenaOperator"]

    def extract(self) -> TaskMetadata:
        task_name = f"{self.operator.dag_id}.{self.operator.task_id}"
        job_facets = {"sql": SqlJobFacet(query=SqlExtractor._normalize_sql(self.operator.query))}

        sql_meta: Optional[SqlMeta] = parse(self.operator.query, "generic", None)
        inputs: List[Dataset] = (
            list(
                filter(
                    None,
                    [
                        self._get_inout_dataset(table.schema or self.operator.database, table.name)
                        for table in sql_meta.in_tables
                    ],
                )
            )
            if sql_meta and sql_meta.in_tables
            else []
        )

        # Athena can output query result to a new table with CTAS query.
        # cf. https://docs.aws.amazon.com/athena/latest/ug/ctas.html
        outputs: List[Dataset] = (
            list(
                filter(
                    None,
                    [
                        self._get_inout_dataset(table.schema or self.operator.database, table.name)
                        for table in sql_meta.out_tables
                    ],
                )
            )
            if sql_meta and sql_meta.out_tables
            else []
        )

        # In addition to CTAS query, it's also possible to specify output location on S3
        # with a mandatory parameter, OutputLocation in ResultConfiguration.
        # cf. https://docs.aws.amazon.com/athena/latest/APIReference/API_ResultConfiguration.html#athena-Type-ResultConfiguration-OutputLocation  # noqa: E501
        #
        # Depending on the query type and the external_location property in the CTAS query,
        # its behavior changes as follows:
        #
        # * Normal SELECT statement
        #   -> The result is put into output_location as files rather than a table.
        #
        # * CTAS statement without external_location (`CREATE TABLE ... AS SELECT ...`)
        #   -> The result is put into output_location as a table,
        #      that is, both metadata files and data files are in there.
        #
        # * CTAS statement with external_location
        #   (`CREATE TABLE ... WITH (external_location='s3://bucket/key') AS SELECT ...`)
        #   -> The result is output as a table, but metadata and data files are
        #      separated into output_location and external_location respectively.
        #
        # For the last case, output_location may be unnecessary as OL's output information,
        # but we keep it as of now since it may be useful for some purpose.
        output_location = self.operator.output_location
        parsed = urlparse(output_location)
        outputs.append(
            Dataset(
                name=parsed.path,
                source=Source(
                    scheme=parsed.scheme,
                    authority=parsed.netloc,
                    connection_url=output_location,
                ),
            )
        )

        return TaskMetadata(
            name=task_name,
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[ds.to_openlineage_dataset() for ds in outputs],
            run_facets={},
            job_facets=job_facets,
        )

    def _get_inout_dataset(self, database, table) -> Optional[Dataset]:
        # Currently, AthenaOperator and AthenaHook don't have a functionality to specify catalog,
        # and it seems to implicitly assume that the default catalog (AwsDataCatalog) is target.
        CATALOG_NAME = "AwsDataCatalog"

        # AthenaHook.get_conn() doesn't return PEP-249 compliant connection object,
        # but Boto3's Athena.Client instead. So this class doesn't inherit from
        # SqlExtractor, which depends on PEP-249 compliant connection.
        client = self.operator.hook.get_conn()
        try:
            table_metadata = client.get_table_metadata(
                CatalogName=CATALOG_NAME, DatabaseName=database, TableName=table
            )
            s3_location = table_metadata["TableMetadata"]["Parameters"]["location"]
            table_schema = DbTableSchema(
                schema_name=database,
                table_name=DbTableMeta(f"{CATALOG_NAME}.{database}.{table}"),
                columns=[
                    DbColumn(name=column["Name"], type=column["Type"], ordinal_position=i)
                    for i, column in enumerate(table_metadata["TableMetadata"]["Columns"])
                ],
            )

            scheme = "awsathena"
            authority = f"athena.{client._client_config.region_name}.amazonaws.com"

            return Dataset.from_table_schema(
                source=Source(
                    scheme=scheme,
                    authority=authority,
                    connection_url=f"{scheme}://{authority}",
                ),
                table_schema=table_schema,
                database_name=CATALOG_NAME,
                data_location=s3_location,
            )

        except Exception as e:
            self.log.error(f"Cannot retrieve table metadata from Athena.Client. {e}", exc_info=True)
            return None



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import os

# Provide empty plugin for older version
from openlineage.airflow.macros import lineage_parent_id, lineage_run_id
from pkg_resources import parse_version

from airflow.plugins_manager import AirflowPlugin
from airflow.version import version as AIRFLOW_VERSION


def _is_disabled():
    try:
        # If the Airflow provider is installed, skip running the openlineage-airflow plugin.
        from airflow.providers.openlineage.plugins.openlineage import (
            OpenLineageProviderPlugin,  # noqa: F401
        )

        return True
    except ImportError:
        pass
    return os.getenv("OPENLINEAGE_DISABLED", "").lower() == "true"


if parse_version(AIRFLOW_VERSION) < parse_version("2.3.0.dev0") or _is_disabled():  # type: ignore

    class OpenLineagePlugin(AirflowPlugin):
        name = "OpenLineagePlugin"
        macros = [lineage_run_id, lineage_parent_id]
else:
    from openlineage.airflow import listener

    # Provide entrypoint airflow plugin that registers listener module
    class OpenLineagePlugin(AirflowPlugin):  # type: ignore
        name = "OpenLineagePlugin"
        listeners = [listener]
        macros = [lineage_run_id, lineage_parent_id]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional
from urllib.parse import urlparse

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.run import Dataset


class S3CopyObjectExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["S3CopyObjectOperator"]

    def extract(self) -> Optional[TaskMetadata]:
        input_object = Dataset(
            namespace=f"s3://{self.operator.source_bucket_name}",
            name=f"s3://{self.operator.source_bucket_name}/{self.operator.source_bucket_key}",
            facets={},
        )

        output_object = Dataset(
            namespace=f"s3://{self.operator.dest_bucket_name}",
            name=f"s3://{self.operator.dest_bucket_name}/{self.operator.dest_bucket_key}",
            facets={},
        )

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=[input_object],
            outputs=[output_object],
        )


class S3FileTransformExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["S3FileTransformOperator"]

    def extract(self) -> Optional[TaskMetadata]:
        input_object = Dataset(
            namespace=f"s3://{urlparse(self.operator.source_s3_key).netloc}",
            name=self.operator.source_s3_key,
            facets={},
        )

        output_object = Dataset(
            namespace=f"s3://{urlparse(self.operator.dest_s3_key).netloc}",
            name=self.operator.dest_s3_key,
            facets={},
        )

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=[input_object],
            outputs=[output_object],
        )

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        pass



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import uuid
from unittest.mock import patch

import pandas as pd
from openlineage.airflow.utils import is_airflow_version_enough

if not is_airflow_version_enough("2.6.0"):
    from airflow.listeners.events import (
        register_task_instance_state_events,
        unregister_task_instance_state_events,
    )
from airflow.models import DAG, BaseOperator, TaskInstance
from airflow.utils.dates import days_ago
from airflow.utils.state import State


class TemplateOperator(BaseOperator):
    template_fields = ["df"]

    def __init__(self, df, *args, **kwargs):
        self.df = df
        super().__init__(*args, **kwargs)

    def execute(self, context):
        return self.df


def render_df():
    return pd.DataFrame({"col": [1, 2]})


@patch("airflow.models.TaskInstance.xcom_push")
@patch("airflow.models.BaseOperator.render_template")
def test_listener_does_not_change_task_instance(render_mock, xcom_push_mock):
    if not is_airflow_version_enough("2.6.0"):
        register_task_instance_state_events()
    render_mock.return_value = render_df()

    dag = DAG(
        "test",
        start_date=days_ago(1),
        user_defined_macros={"render_df": render_df},
        params={"df": render_df()},
    )
    t = TemplateOperator(task_id="template_op", dag=dag, do_xcom_push=True, df=dag.param("df"))
    run_id = str(uuid.uuid1())
    dag.create_dagrun(state=State.NONE, run_id=run_id)
    ti = TaskInstance(t, run_id=run_id)
    ti.check_and_change_state_before_execution()  # make listener hook on running event
    ti._run_raw_task()

    if not is_airflow_version_enough("2.6.0"):
        # we need to unregister hooks not to break BigQuery E2E tests
        unregister_task_instance_state_events()

    # check if task returns the same DataFrame
    pd.testing.assert_frame_equal(xcom_push_mock.call_args[1]["value"], render_df())

    # check if render_template method always get the same unrendered field
    assert not isinstance(render_mock.call_args[0][0], pd.DataFrame)


@patch("openlineage.airflow.listener.getboolean")
@patch("openlineage.airflow.listener.is_airflow_version_enough")
@patch("openlineage.airflow.listener.execute_in_thread")
def test_listener_chooses_direct_execution_when_env_variable(
    execute_in_thread, is_airflow_version_enough, getboolean
):
    getboolean.return_value = True
    is_airflow_version_enough.return_value = False
    is_called = False

    def call():
        nonlocal is_called
        is_called = True

    from openlineage.airflow.listener import execute

    execute(call)
    assert is_called


@patch("openlineage.airflow.utils.getboolean")
@patch("openlineage.airflow.listener.is_airflow_version_enough")
@patch("openlineage.airflow.listener.execute_in_thread")
def test_listener_chooses_thread_execution(execute_in_thread, is_airflow_version_enough, getboolean):
    getboolean.return_value = False
    is_airflow_version_enough.return_value = False
    is_called = False

    def call():
        nonlocal is_called
        is_called = True

    from openlineage.airflow.listener import execute

    execute(call)
    assert not is_called



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
from contextlib import closing
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple

from openlineage.common.dataset import Dataset, Source
from openlineage.common.models import DbColumn, DbTableSchema
from openlineage.common.sql import DbTableMeta

if TYPE_CHECKING:
    from airflow.hooks.base import BaseHook


logger = logging.getLogger(__name__)

_TABLE_SCHEMA = 0
_TABLE_NAME = 1
_COLUMN_NAME = 2
_ORDINAL_POSITION = 3
# Use 'udt_name' which is the underlying type of column
# (ex: int4, timestamp, varchar, etc)
_UDT_NAME = 4
# Database is optional as 5th column
_TABLE_DATABASE = 5

TablesHierarchy = Dict[str, Dict[str, List[str]]]


def execute_query_on_hook(hook: "BaseHook", query: str) -> Iterator[tuple]:
    with closing(hook.get_conn()) as conn:
        with closing(conn.cursor()) as cursor:
            return cursor.execute(query).fetchall()


def get_table_schemas(
    hook: "BaseHook",
    source: Source,
    database: str,
    in_query: Optional[str],
    out_query: Optional[str],
) -> Tuple[List[Dataset], ...]:
    """
    This function queries database for table schemas using provided hook.
    Responsibility to provide queries for this function is on particular extractors.
    If query for input or output table isn't provided, the query is skipped.
    """
    query_schemas = []

    # Do not query if we did not get both queries
    if not in_query and not out_query:
        return [], []

    with closing(hook.get_conn()) as conn:
        with closing(conn.cursor()) as cursor:
            for query in [in_query, out_query]:
                if query:
                    cursor.execute(query)
                    query_schemas.append(parse_query_result(cursor))
                else:
                    query_schemas.append([])
    return tuple(
        [
            [
                Dataset.from_table_schema(source=source, table_schema=schema, database_name=db or database)
                for schema, db in schemas
            ]
            for schemas in query_schemas
        ]
    )


def parse_query_result(cursor) -> List[Tuple[DbTableSchema, str]]:
    schemas: Dict = {}
    for row in cursor.fetchall():
        table_schema_name: str = row[_TABLE_SCHEMA]
        table_name: DbTableMeta = DbTableMeta(row[_TABLE_NAME])
        table_column: DbColumn = DbColumn(
            name=row[_COLUMN_NAME],
            type=row[_UDT_NAME],
            ordinal_position=row[_ORDINAL_POSITION],
        )
        try:
            table_database = row[_TABLE_DATABASE]
        except IndexError:
            table_database = None

        # Attempt to get table schema
        table_key = ".".join(filter(None, [table_database, table_schema_name, table_name.name]))
        # table_key: str = f"{table_schema_name}.{table_name}"
        table_schema: Optional[DbTableSchema]
        table_schema, _ = schemas.get(table_key) or (None, None)

        if table_schema:
            # Add column to existing table schema.
            schemas[table_key][0].columns.append(table_column)
        else:
            # Create new table schema with column.
            schemas[table_key] = (
                DbTableSchema(
                    schema_name=table_schema_name,
                    table_name=table_name,
                    columns=[table_column],
                ),
                table_database,
            )
    return list(schemas.values())


def create_information_schema_query(
    columns: List[str],
    information_schema_table_name: str,
    tables_hierarchy: TablesHierarchy,
    uppercase_names: bool = False,
    allow_trailing_semicolon: bool = True,
) -> str:
    """
    This function creates query for getting table schemas from information schema.
    """
    sqls = []
    for db, schema_mapping in tables_hierarchy.items():
        filter_clauses = create_filter_clauses(schema_mapping, uppercase_names)
        source = information_schema_table_name
        if db:
            source = f"{db.upper() if uppercase_names else db}." f"{source}"
        sqls.append(f"SELECT {', '.join(columns)} " f"FROM {source} " f"WHERE {' OR '.join(filter_clauses)}")
    sql = " UNION ALL ".join(sqls)

    # For some databases such as Trino, trailing semicolon can cause a syntax error.
    if allow_trailing_semicolon:
        sql += ";"

    return sql


def create_filter_clauses(schema_mapping, uppercase_names: bool = False) -> List[str]:
    filter_clauses = []
    for schema, tables in schema_mapping.items():
        table_names = ",".join(
            map(
                lambda name: f"'{name.upper() if uppercase_names else name}'",
                tables,
            )
        )
        if schema:
            filter_clauses.append(
                f"( table_schema = '{schema.upper() if uppercase_names else schema}' "
                f"AND table_name IN ({table_names}) )"
            )
        else:
            filter_clauses.append(f"( table_name IN ({table_names}) )")
    return filter_clauses



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
import os
from unittest import mock
from unittest.mock import MagicMock, patch

from openlineage.airflow.adapter import OpenLineageAdapter


@patch.dict(os.environ, {"MARQUEZ_URL": "http://marquez:5000", "MARQUEZ_API_KEY": "api-key"})
def test_create_client_from_marquez_url():
    client = OpenLineageAdapter().get_or_create_openlineage_client()
    assert client.transport.url == "http://marquez:5000"


@patch.dict(
    os.environ,
    {"OPENLINEAGE_URL": "http://ol-api:5000", "OPENLINEAGE_API_KEY": "api-key"},
)
def test_create_client_from_ol_env():
    client = OpenLineageAdapter().get_or_create_openlineage_client()
    assert client.transport.url == "http://ol-api:5000"


def test_setting_ol_adapter_log_level() -> None:
    # DEBUG level set for `openlineage` logger in tests setup
    default_log_level = logging.DEBUG
    # without environment variable
    OpenLineageAdapter()
    parent_logger = logging.getLogger("openlineage.airflow")
    logger = logging.getLogger("openlineage.airflow.adapter")
    assert parent_logger.getEffectiveLevel() == default_log_level
    assert logger.getEffectiveLevel() == default_log_level
    with patch.dict(os.environ, {"OPENLINEAGE_AIRFLOW_LOGGING": "CRITICAL"}):
        assert parent_logger.getEffectiveLevel() == default_log_level
        assert logger.getEffectiveLevel() == default_log_level
        OpenLineageAdapter()
        assert parent_logger.getEffectiveLevel() == logging.CRITICAL
        assert logger.getEffectiveLevel() == logging.CRITICAL


@patch("openlineage.airflow.adapter.OpenLineageAdapter.get_or_create_openlineage_client")
@patch("openlineage.airflow.adapter.redact_with_exclusions")
@patch("openlineage.airflow.adapter.Stats.incr")
@patch("openlineage.airflow.adapter.Stats.timer")
def test_openlineage_adapter_stats_emit_success(
    mock_stats_timer, mock_stats_incr, mock_redact, mock_get_client
):
    adapter = OpenLineageAdapter()

    adapter.emit(MagicMock())

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@patch("openlineage.airflow.adapter.OpenLineageAdapter.get_or_create_openlineage_client")
@patch("openlineage.airflow.adapter.redact_with_exclusions")
@patch("openlineage.airflow.adapter.Stats.incr")
@patch("openlineage.airflow.adapter.Stats.timer")
def test_openlineage_adapter_stats_emit_failed(
    mock_stats_timer, mock_stats_incr, mock_redact, mock_get_client
):
    adapter = OpenLineageAdapter()
    mock_get_client.return_value.emit.side_effect = Exception()

    adapter.emit(MagicMock())

    mock_stats_timer.assert_called_with("ol.emit.attempts")
    mock_stats_incr.assert_has_calls([mock.call("ol.emit.failed")])



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, List

import boto3
from openlineage.airflow.extractors.postgres_extractor import PostgresExtractor

if TYPE_CHECKING:
    from airflow.hooks.base import BaseHook


class RedshiftSQLExtractor(PostgresExtractor):
    _whitelist_query_params: List[str] = ["cluster_identifier", "region"]

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["RedshiftSQLOperator"]

    @property
    def dialect(self):
        return "redshift"

    def _get_scheme(self) -> str:
        return "redshift"

    def _get_hook(self) -> "BaseHook":
        from airflow.providers.amazon.aws.hooks.redshift_sql import RedshiftSQLHook

        return RedshiftSQLHook(
            redshift_conn_id=self.operator.redshift_conn_id,
        )

    def _get_authority(self) -> str:
        # If IAM enabled disregard hostname, use info from extras
        # Important: RedshiftSQLOperator does not handle
        # Redshift Serverless connections yet neither Airflow AWS provider does.
        extras = self.conn.extra_dejson
        if extras.get("iam") is True:
            host = extras.get("cluster_identifier")
            region = extras.get("region")
            if not region:
                profile = extras.get("profile", None)
                session = boto3.Session(profile_name=profile)
                region = session.region_name
            port = extras.get("port", 5439)
            identifier = f"{host}.{region}"
        elif not self.conn.host:
            raise ValueError("Missing host in connection since there" "s no IAM setting")
        else:
            identifier = self._get_cluster_identifier_from_hostname(self.conn.host)
            port = self.conn.port or 5439
        return f"{identifier}:{port}"

    def _get_cluster_identifier_from_hostname(self, hostname: str) -> str:
        parts = hostname.split(".")
        if "amazonaws.com" in hostname and len(parts) == 6:
            return f"{parts[0]}.{parts[2]}"
        else:
            self.log.warning(
                """Could not parse identifier from hostname '%s'.
            You are probably using IP to connect to Redshift cluster.
            Expected format: 'cluster_identifier.id.region_name.redshift.amazonaws.com'
            Falling back to whole hostname.""",
                hostname,
            )
            return hostname



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import copy
import datetime
import logging
import threading
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import TYPE_CHECKING, Callable, Optional

from openlineage.airflow.adapter import OpenLineageAdapter
from openlineage.airflow.extractors import ExtractorManager
from openlineage.airflow.utils import (
    DagUtils,
    get_airflow_run_facet,
    get_custom_facets,
    get_dagrun_start_end,
    get_job_name,
    get_task_location,
    getboolean,
    is_airflow_version_enough,
)

from airflow.listeners import hookimpl

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.models import BaseOperator, DagRun, TaskInstance


class TaskHolder:
    """Class that stores run data - run_id and task in-memory. This is needed because Airflow
    does not always pass all runtime info to on_task_instance_success and
    on_task_instance_failed that is needed to emit events. This is not a big problem since
    we're only running on worker - in separate process that is always spawned (or forked) on
    execution, just like old PHP runtime model.
    """

    def __init__(self):
        self.run_data = {}

    def set_task(self, task_instance: "TaskInstance"):
        self.run_data[self._pk(task_instance)] = task_instance.task

    def get_task(self, task_instance: "TaskInstance") -> Optional["BaseOperator"]:
        return self.run_data.get(self._pk(task_instance))

    @staticmethod
    def _pk(ti: "TaskInstance"):
        return ti.dag_id + ti.task_id + ti.run_id


log = logging.getLogger(__name__)
# TODO: move task instance runs to executor
executor: Optional[Executor] = None


def execute_in_thread(target: Callable, kwargs=None):
    if kwargs is None:
        kwargs = {}
    thread = threading.Thread(target=target, kwargs=kwargs, daemon=True)
    thread.start()

    # Join, but ignore checking if thread stopped. If it did, then we shouldn't do anything.
    # This basically gives this thread 5 seconds to complete work, then it can be killed,
    # as daemon=True. We don't want to deadlock Airflow if our code hangs.

    # This will hang if this timeouts, and extractor is running non-daemon thread inside,
    # since it will never be cleaned up. Ex. SnowflakeOperator
    thread.join(timeout=10)


task_holder = TaskHolder()
extractor_manager = ExtractorManager()
adapter = OpenLineageAdapter()


def direct_execution():
    return is_airflow_version_enough("2.6.0") or getboolean(
        "OPENLINEAGE_AIRFLOW_ENABLE_DIRECT_EXECUTION", False
    )


def execute(_callable):
    try:
        if direct_execution():
            _callable()
        else:
            execute_in_thread(_callable)
    except Exception:
        # Make sure we're not failing task, even for things we think can't happen
        log.exception("Failed to emit OpenLineage event due to exception")


@hookimpl
def on_task_instance_running(previous_state, task_instance: "TaskInstance", session: "Session"):
    if not hasattr(task_instance, "task"):
        log.warning(
            "No task set for TI object task_id: %s - dag_id: %s - run_id %s",
            task_instance.task_id,
            task_instance.dag_id,
            task_instance.run_id,
        )
        return

    log.debug("OpenLineage listener got notification about task instance start")
    dagrun = task_instance.dag_run

    def on_running():
        nonlocal task_instance
        ti = copy.deepcopy(task_instance)
        ti.render_templates()

        task = ti.task
        dag = task.dag
        task_holder.set_task(ti)
        # that's a workaround to detect task running from deferred state
        # we return here because Airflow 2.3 needs task from deferred state
        if ti.next_method is not None:
            return
        parent_run_id = adapter.build_dag_run_id(task.dag.dag_id, dagrun.run_id)

        task_uuid = OpenLineageAdapter.build_task_instance_run_id(
            task.task_id, ti.execution_date, ti.try_number
        )

        task_metadata = extractor_manager.extract_metadata(dagrun, task, task_uuid=task_uuid)

        ti_start_time = ti.start_date if ti.start_date else datetime.datetime.now()
        start, end = get_dagrun_start_end(dagrun=dagrun, dag=dag)

        adapter.start_task(
            run_id=task_uuid,
            job_name=get_job_name(task),
            job_description=dag.description,
            event_time=DagUtils.get_start_time(ti_start_time),
            parent_job_name=dag.dag_id,
            parent_run_id=parent_run_id,
            code_location=get_task_location(task),
            nominal_start_time=DagUtils.get_start_time(start),
            nominal_end_time=DagUtils.to_iso_8601(end),
            owners=dag.owner.split(", "),
            task=task_metadata,
            run_facets={
                **task_metadata.run_facets,
                **get_custom_facets(dagrun, task, dagrun.external_trigger, ti),
                **get_airflow_run_facet(dagrun, dag, ti, task, task_uuid),
            },
        )

    execute(on_running)


@hookimpl
def on_task_instance_success(previous_state, task_instance: "TaskInstance", session):
    log.debug("OpenLineage listener got notification about task instance success")
    task = task_holder.get_task(task_instance) or task_instance.task

    dagrun = task_instance.dag_run

    task_uuid = OpenLineageAdapter.build_task_instance_run_id(
        task.task_id, task_instance.execution_date, task_instance.try_number - 1
    )

    def on_success():
        task_metadata = extractor_manager.extract_metadata(
            dagrun, task, complete=True, task_instance=task_instance
        )
        adapter.complete_task(
            run_id=task_uuid,
            job_name=get_job_name(task),
            end_time=DagUtils.to_iso_8601(task_instance.end_date),
            task=task_metadata,
        )

    execute(on_success)


@hookimpl
def on_task_instance_failed(previous_state, task_instance: "TaskInstance", session):
    log.debug("OpenLineage listener got notification about task instance failure")
    task = task_holder.get_task(task_instance) or task_instance.task

    dagrun = task_instance.dag_run

    task_uuid = OpenLineageAdapter.build_task_instance_run_id(
        task.task_id, task_instance.execution_date, task_instance.try_number - 1
    )

    def on_failure():
        task_metadata = extractor_manager.extract_metadata(
            dagrun, task, complete=True, task_instance=task_instance
        )

        end_date = task_instance.end_date if task_instance.end_date else datetime.datetime.now()

        adapter.fail_task(
            run_id=task_uuid,
            job_name=get_job_name(task),
            end_time=DagUtils.to_iso_8601(end_date),
            task=task_metadata,
        )

    execute(on_failure)


@hookimpl
def on_starting(component):
    global executor
    executor = ThreadPoolExecutor(max_workers=8, thread_name_prefix="openlineage_")


@hookimpl
def before_stopping(component):
    executor.shutdown(wait=False)


@hookimpl
def on_dag_run_running(dag_run: "DagRun", msg: str):
    if not executor:
        log.error("Executor have not started before `on_dag_run_running`")
        return
    start, end = get_dagrun_start_end(dag_run, dag_run.dag)
    executor.submit(
        adapter.dag_started,
        dag_run=dag_run,
        msg=msg,
        nominal_start_time=DagUtils.get_start_time(start),
        nominal_end_time=DagUtils.to_iso_8601(end),
    )


@hookimpl
def on_dag_run_success(dag_run: "DagRun", msg: str):
    if not executor:
        log.error("Executor have not started before `on_dag_run_success`")
        return
    executor.submit(adapter.dag_success, dag_run=dag_run, msg=msg)


@hookimpl
def on_dag_run_failed(dag_run: "DagRun", msg: str):
    if not executor:
        log.error("Executor have not started before `on_dag_run_failed`")
        return
    executor.submit(adapter.dag_failed, dag_run=dag_run, msg=msg)



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

sql_extractors = {
    "bigquery": "BigQueryExtractor",
    "mysql": "MysqlExtractor",
    "postgres": "PostgresExtractor",
    "redshift": "RedshiftExtractor",
    "snowflake": "SnowflakeExtractor",
}


def get_sql_execute_query_extractor(super_):
    class SQLExecuteQueryExtractor(super_):
        pass

    return SQLExecuteQueryExtractor
